from __future__ import annotations

import contextlib
import functools
import http
import socket
import ssl
import warnings
from typing import TYPE_CHECKING, NamedTuple
from urllib.request import urlopen

from cryptography import x509

from aia_chaser.constants import DEFAULT_URLOPEN_TIMEOUT, DOWNLOAD_CACHE_SIZE
from aia_chaser.exceptions import (
    AiaChaseExhaustedError,
    CertificateDownloadError,
    CertificateParseError,
    MissingPeerCertificateError,
    NoValidAiaCaUrlError,
    RootCertificateNotFoundError,
)
from aia_chaser.utils.cert_utils import (
    certificates_to_der,
    extract_aia_information,
    load_ssl_ca_certificates,
)
from aia_chaser.utils.url import extract_host_port_from_url
from aia_chaser.verify import VerifyCertificatesConfig, verify_certificate_chain


if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence


__all__ = ["AiaChaser"]


class AiaChaser:
    """Authority Information Access (AIA) Chaser.

    AIA is part of the X509 standard in RFC 5280. It's objective
    is pointing the client towards an endpoint from which the
    signing certificate can be obtained even if the server does
    not provide the intermediate certificates as part of the
    TLS handshake.

    The chaser object can be later used to generate SSL context
    to access specific hosts after all the intermediate certificates
    have been resolved.

    Args:
        context: Context used internally to request the host's certificate
            during AIA chasing operations. Its loaded CA certificates at
            the time of crating the `AiaChaser` are considered the trusted
            root CAs. If not given a new SSLContext is created with the
            default certificates.
    """

    def __init__(self, context: ssl.SSLContext | None = None) -> None:
        self._context = context or ssl.SSLContext()

        # Load trusted certificates
        trusted_cert = load_ssl_ca_certificates(
            self._context,
            force_load=context is None,
        )

        self._trusted = {ca_cert.subject: ca_cert for ca_cert in trusted_cert}

    def aia_chase_cert(
        self,
        certificate: x509.Certificate,
    ) -> Iterator[x509.Certificate]:
        """Chase AIA CA information from the provided certificate.

        Args:
            certificate: Start AIA chasing from this certificate.

        Yields:
            The certificates from the certificate chain of
                `certificate`. The first is the provided
                certificate and the last is the root or a
                trusted CA.
        """
        cert = certificate
        while True:
            cert_info = _extract_aia_info(cert)

            # Already trusted & self-signed
            #
            # Running until the end of some chains end up in "ValiCert Class 2
            # Policy Validation Authority", which are not in the system's
            # trusted database:
            # https://security.stackexchange.com/questions/65508/what-is-the-deal-with-valicert-ssl-root-certificates
            #
            # This may happen with other certificate chains thereby we stop
            # the chasing once we find a certificate that we already trust.
            if cert_info.subject in self._trusted:
                yield self._trusted[cert_info.subject]
                break
            if cert_info.subject == cert_info.issuer:
                if cert_info.issuer not in self._trusted:
                    raise RootCertificateNotFoundError(
                        subject=cert_info.issuer.rfc4514_string(),
                    )
                yield self._trusted[cert_info.issuer]
                break

            # Yield and continue AIA chasing
            yield cert

            # No more intermediate CAs, issuer must be trusted
            if not cert_info.aia_ca_issuers:
                if cert_info.issuer not in self._trusted:
                    raise RootCertificateNotFoundError(
                        subject=cert_info.issuer.rfc4514_string(),
                    )
                yield self._trusted[cert_info.issuer]
                break

            cert = _try_download_certificate(cert_info.aia_ca_issuers)

    def aia_chase(
        self,
        host: str,
        port: int = 443,
    ) -> Iterator[x509.Certificate]:
        """Chase AIA CA information starting from the `host`'s certificate.

        Args:
            host: Host to get the initial certificate.
            port: Port on host to connect and retrieve the initial
                certificate.

        Yields:
            The certificates from the certificate chain of the
                host's certificate. The first is the host's
                certificate and the last is the root or a
                trusted CA.
        """
        cert = self.fetch_host_cert(host=host, port=port)
        yield from self.aia_chase_cert(certificate=cert)

    def fetch_host_cert(self, host: str, port: int = 443) -> x509.Certificate:
        """Get the host, port pair certificate.

        Args:
            host: Host to retrieve the certificate for.
            port: Port on host to connect and retrieve the certificate.

        Returns:
            The certificate of the (host, port) pair.

        Raises:
            MissingPeerCertificateError: If it is not possible to retrieve
                the certificate of the `host`.
        """
        with contextlib.ExitStack() as ctx:
            sock = ctx.enter_context(socket.create_connection((host, port)))
            s_sock = ctx.enter_context(
                self._context.wrap_socket(
                    sock,
                    server_hostname=host,
                ),
            )

            # Get initial, possibly incomplete, chain from peer once the
            # functionality is supported by ssl module
            # https://www.openssl.org/docs/man1.0.2/man3/SSL_get_peer_cert_chain.html
            der_cert = s_sock.getpeercert(True)  # noqa: FBT003
            if der_cert is not None:
                return x509.load_der_x509_certificate(der_cert)

        raise MissingPeerCertificateError(host=host, port=port)

    def fetch_cert_chain_for_host(
        self,
        host: str,
        port: int = 443,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> list[x509.Certificate]:
        """Fetch the certificate chain for a given host.

        Retrieves the certificate chain for a specified host and port.
        Optionally verifies the chain against a trusted certificate store.

        Args:
            host: The hostname to fetch the certificate chain for.
            port: The port to connect to. Defaults to 443.
            verify_config: Configuration for verifying the certificate chain.
                If None, a default configuration is used.
            verify: Whether to verify the certificate chain. Defaults to True.

        Returns:
            The fetched certificate chain, optionally verified.

        Raises:
            exceptions.CertificateChainError: If the certificate chain fails
                verification.
            exceptions.MissingPeerCertificateError: If it is not possible to
                retrieve the certificate of the `host`.
        """
        certificates = list(self.aia_chase(host, port))
        if verify:
            verify_certificate_chain(
                certificates=certificates,
                trusted=self._trusted,
                config=verify_config or VerifyCertificatesConfig(),
            )
        return certificates

    def fetch_ca_chain_for_host(
        self,
        host: str,
        port: int = 443,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> list[x509.Certificate]:
        """Fetch the CA certificate chain for a given host.

        Same as
        [`fetch_cert_chain_for_host`][aia_chaser.AiaChaser.fetch_cert_chain_for_host]
        excluding the host's certificate.

        Args:
            host: The hostname to fetch the CA certificate chain for.
            port: The port to connect to. Defaults to 443.
            verify_config: Configuration for verifying the certificate chain.
                If None, a default configuration is used.
            verify: Whether to verify the CA certificate chain. Defaults to True.

        Returns:
            The fetched CA certificate chain (host excluded), optionally verified.

        Raises:
            exceptions.CertificateChainError: If the certificate chain fails
                verification.
            exceptions.MissingPeerCertificateError: If it is not possible to
                retrieve the certificate of the `host`.
        """
        return self.fetch_cert_chain_for_host(
            host,
            port,
            verify_config=verify_config,
            verify=verify,
        )[:1]

    def fetch_cert_chain_for_url(
        self,
        url_string: str,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> list[x509.Certificate]:
        """Fetch the certificate chain for a given host.

        Same as `fetch_cert_chain_for_host` but the host name and port
        are obtained from the `url_string`

        Args:
            url_string: URL to fetch the certificate chain for.
            verify_config: Configuration for verifying the certificate chain.
                If None, a default configuration is used.
            verify: Whether to verify the certificate chain. Defaults to True.

        Returns:
            The fetched certificate chain, optionally verified.

        Raises:
            exceptions.CertificateChainError: If the certificate chain fails
                verification.
            exceptions.MissingPeerCertificateError: If it is not possible to
                retrieve the certificate of the host.
        """
        host, port = extract_host_port_from_url(url_string)
        return self.fetch_cert_chain_for_host(
            host=host,
            port=port,
            verify_config=verify_config,
            verify=verify,
        )

    def fetch_ca_chain_for_url(
        self,
        url_string: str,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> list[x509.Certificate]:
        """Fetch the CA certificate chain for a given host.

        Same as `fetch_cert_chain_for_url` excluding the host's
        certificate.

        Args:
            url_string: URL to fetch the CA certificate chain for.
            verify_config: Configuration for verifying the certificate chain.
                If None, a default configuration is used.
            verify: Whether to verify the CA certificate chain. Defaults to True.

        Returns:
            The fetched CA certificate chain, optionally verified.

        Raises:
            exceptions.CertificateChainError: If the certificate chain fails
                verification.
            exceptions.MissingPeerCertificateError: If it is not possible to
                retrieve the certificate of the host.
        """
        return self.fetch_cert_chain_for_url(
            url_string=url_string,
            verify_config=verify_config,
            verify=verify,
        )[:1]

    def make_ssl_context_for_host(
        self,
        host: str,
        port: int = 443,
        purpose: ssl.Purpose = ssl.Purpose.SERVER_AUTH,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> ssl.SSLContext:
        """Create a new SSL context and add certificate chain for host.

        See Also:
            fetch_ca_chain_for_host: Method used to retrieve and optionally
                verify the CA chain to add to the SSL context.
        """
        context = ssl.create_default_context(purpose=purpose)
        self.add_host_ca_chain_to_context(
            context,
            host,
            port,
            verify_config=verify_config,
            verify=verify,
        )
        return context

    def make_ssl_context_for_url(
        self,
        url_string: str,
        purpose: ssl.Purpose = ssl.Purpose.SERVER_AUTH,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> ssl.SSLContext:
        """Create a new SSL context and add certificate chain for URL.

        See Also:
            fetch_ca_chain_for_url: Method used to retrieve and optionally
                verify the CA chain to add to the SSL context.
        """
        context = ssl.create_default_context(purpose=purpose)
        self.add_url_ca_chain_to_context(
            context,
            url_string=url_string,
            verify_config=verify_config,
            verify=verify,
        )
        return context

    def add_host_ca_chain_to_context(
        self,
        context: ssl.SSLContext,
        host: str,
        port: int = 443,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> None:
        """Add host CA chain to SSL context.

        See Also:
            fetch_ca_chain_for_host: Method used to retrieve and optionally
                verify the CA chain to add to the SSL context.
        """
        certificates = self.fetch_ca_chain_for_host(
            host,
            port,
            verify_config=verify_config,
            verify=verify,
        )
        context.load_verify_locations(cadata=certificates_to_der(certificates))

    def add_url_ca_chain_to_context(
        self,
        context: ssl.SSLContext,
        url_string: str,
        verify_config: VerifyCertificatesConfig | None = None,
        *,
        verify: bool = True,
    ) -> None:
        """Add CA chain to URL's host to SSL context.

        See Also:
            fetch_ca_chain_for_host: Method used to retrieve and optionally
                verify the CA chain to add to the SSL context.
        """
        certificates = self.fetch_ca_chain_for_url(
            url_string,
            verify_config=verify_config,
            verify=verify,
        )
        context.load_verify_locations(cadata=certificates_to_der(certificates))


def _try_download_certificate(urls: Sequence[str]) -> x509.Certificate:
    http_urls = [url for url in urls if url.lower().startswith(("http:", "https:"))]
    if not http_urls:
        raise NoValidAiaCaUrlError(urls=urls)

    errors = []
    for http_url in http_urls:
        try:
            return _download_certificate(http_url)
        except (  # noqa: PERF203
            CertificateDownloadError,
            CertificateParseError,
        ) as err:
            errors.append(err)

    raise AiaChaseExhaustedError(errors=errors)


@functools.lru_cache(maxsize=DOWNLOAD_CACHE_SIZE)
def _download_certificate(url_string: str) -> x509.Certificate:
    if url_string.startswith("https:"):
        warnings.warn(
            "Trying to download an intermediate CA certificate using HTTPS",
            category=UserWarning,
            stacklevel=2,
        )
    elif not url_string.startswith("http:"):
        raise CertificateDownloadError(
            message="URL scheme must be http or https",
            url_string=url_string,
        )

    with urlopen(  # noqa: S310
        url_string,
        timeout=DEFAULT_URLOPEN_TIMEOUT,
    ) as response:
        if response.status != http.HTTPStatus.OK:
            raise CertificateDownloadError(
                message=f"could not download {url_string}",
                url_string=url_string,
            )

        try:
            return _try_parse_certificate(response.read())
        except CertificateParseError as err:
            raise CertificateDownloadError(
                message=str(err),
                url_string=url_string,
            ) from None


def _try_parse_certificate(data: bytes) -> x509.Certificate:
    parse_fns = [x509.load_der_x509_certificate, x509.load_pem_x509_certificate]
    exceptions = []
    for parse_fn in parse_fns:
        try:
            return parse_fn(data)
        except (ValueError, TypeError) as err:  # noqa: PERF203
            exceptions.append(err)

    raise CertificateParseError(reasons=[str(err) for err in exceptions])


class _CertificateAiaInfo(NamedTuple):
    """Simpler format to work with  certificate info."""

    subject: x509.Name
    issuer: x509.Name
    aia_ca_issuers: list[str]
    aia_ocsp_urls: list[str]


def _extract_aia_info(x509_certificate: x509.Certificate) -> _CertificateAiaInfo:
    aia_info = extract_aia_information(x509_certificate)
    return _CertificateAiaInfo(
        subject=x509_certificate.subject,
        issuer=x509_certificate.issuer,
        aia_ca_issuers=aia_info.ca_issuers,
        aia_ocsp_urls=aia_info.ocsp_urls,
    )



from __future__ import annotations

from urllib.parse import urlsplit

from aia_chaser.constants import DEFAULT_SCHEME_PORT


def extract_host_port_from_url(url_string: str) -> tuple[str, int]:
    """Extract host and port from a URL string.

    If the port is not explicitly specified it will be inferred from the
    scheme.

    Args:
        url_string: URL from which to extract the host.

    Returns:
        The host (netloc) of `url_string` and, if present, the port.

    Raises:
        ValueError: If the host or port cannot be extracted from the given
            URL. It may happen with a seemingly correct URL if
            it is missing the scheme component.
    """
    url_split = urlsplit(url_string)
    if not url_split.netloc:
        raise ValueError("cannot split host from URL, missing scheme?")

    scheme = url_split.scheme.lower()
    port = url_split.port

    if port is None:
        if not scheme:
            raise ValueError("URL has no port nor scheme to infer it")
        if scheme not in DEFAULT_SCHEME_PORT:
            raise ValueError(f"default port for scheme '{scheme}' is not known")

        port = DEFAULT_SCHEME_PORT[scheme]

    return url_split.netloc, port



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import datetime
    from collections.abc import Sequence

    from cryptography.x509.ocsp import OCSPResponseStatus


class AiaChaserError(Exception):
    """Base exception for aia_chaser errors."""


class MissingPeerCertificateError(AiaChaserError):
    """There is no certificate for the peer on the other end."""

    def __init__(self, host: str, port: int) -> None:
        super().__init__(f"could not get certificate for host {host} on port {port}")

        self.host = host
        self.port = port


class CertificateDownloadError(AiaChaserError):
    """Error downloading a certificate."""

    def __init__(
        self,
        message: str,
        url_string: str,
        content_type: str | None = None,
    ) -> None:
        super().__init__(message)
        self.url_string = url_string
        self.content_type = content_type


class CertificateParseError(AiaChaserError):
    """Error parsing a certificate."""

    def __init__(self, reasons: Sequence[str]) -> None:
        super().__init__("could not parse certificate")
        self.reasons = reasons


class NoValidAiaCaUrlError(AiaChaserError):
    """None of the AIA CA URLs is a valid http or https URL.

    Args:
        urls: URLs that were provided to fetch the CA.
    """

    def __init__(self, urls: Sequence[str]) -> None:
        super().__init__(
            f"at least one of the AIA CA urls ({urls}) must be an http or https url",
        )
        self.urls = urls


class AiaChaseExhaustedError(AiaChaserError):
    """Failed to retrieve the issuing CA after exhausting all AIA URLs.

    Arguments:
        errors: A collection of errors encountered while attempting to
            fetch the CA.

    Attributes:
        errors: A collection of errors encountered while attempting to
            fetch the CA.
    """

    def __init__(
        self,
        errors: Sequence[CertificateDownloadError | CertificateParseError],
    ) -> None:
        super().__init__("AIA chasing failed after exhausting all CA urls")
        self.errors = errors


class CertificateVerificationError(AiaChaserError):
    """Base exception for certificate verification errors."""


class CertificateIssuerNameError(CertificateVerificationError):
    """Certificate issuer name does not match issuer's subject.

    Args:
        cert_issuer: Certificate's issuer subject.
        issuer_subject: Issuer's subject.
    """

    def __init__(self, cert_issuer: str, issuer_subject: str) -> None:
        super().__init__(
            f"certificate's issuer name ({cert_issuer}) does not"
            f" match issuer's subject ({issuer_subject})",
        )
        self.cert_issuer = cert_issuer
        self.issuer_subject = issuer_subject


class CertificateKeyTypeError(CertificateVerificationError):
    """Unsupported key type used to sign a certificate.

    Args:
        reason: Message explaining the cause of the error.
    """

    def __init__(self, reason: str) -> None:
        msg = "certificate does not provide a supported public key"
        if reason:
            msg += f" [{reason}]"

        super().__init__(msg)


class CertificateSignatureError(CertificateVerificationError):
    """Issuer's certificate did not sign the certificate.

    Args:
        certificate: Certificate whose signature is being validated.
        issuer: Certificate that should have signed the `certificate`.
    """

    def __init__(self, certificate: str, issuer: str) -> None:
        super().__init__(f"issuer ({issuer}) did not sign certificate ({certificate})")
        self.certificate_name = certificate
        self.issuer_name = issuer


class CertificateIssuerNotTrustedError(CertificateVerificationError):
    """Issuer's certificate cannot be used to verify signature."""

    def __init__(self, issuer: str) -> None:
        super().__init__(
            f"issuer ({issuer}) not found among the trusted certificates",
        )
        self.issuer_name = issuer


class CertificateExpiredError(CertificateVerificationError):
    """Certificate outside its validity period.

    Args:
        not_valid_before: Certificate is not valid before this time.
        not_valid_after: Certificate is not valid after this time.
        verification_time: When the certificate was validated and
            validation failed because it's outside the validity period.
    """

    def __init__(
        self,
        not_valid_before: datetime.datetime,
        not_valid_after: datetime.datetime,
        verification_time: datetime.datetime,
    ) -> None:
        super().__init__(
            "certificate outside of validity period"
            f" [{not_valid_before}, {not_valid_after}]"
            f" using verification time {verification_time}",
        )
        self.not_valid_before = not_valid_after
        self.not_valid_after = not_valid_after
        self.verification_time = verification_time


class CertificateTimeZoneError(CertificateVerificationError):
    """Cannot compare offset-aware and offset-naive times."""

    def __init__(self, message: str | None = None) -> None:
        message = message or (
            "Using offset-naive time is disallowed set the offset with"
            " `datetime.replace(tzinfo=<timezone>) or equivalent"
        )
        super().__init__(message)


class RootCertificateNotFoundError(CertificateVerificationError):
    """Root certificate not in trusted database.

    Args:
        subject: Subject of the certificate not found among the
            trusted certificates.
    """

    def __init__(self, subject: str) -> None:
        super().__init__(
            f"root certificate with subject '{subject}' not in trusted database",
        )


class CertificateFingerprintError(CertificateVerificationError):
    """Certificate fingerprint does not match trusted fingerprint."""

    def __init__(self, fingerprint: bytes, trusted_fingerprint: bytes) -> None:
        super().__init__(
            "certificate fingerprint does not match fingerprint of"
            " certificate in trusted database",
        )
        self.fingerprint = fingerprint
        self.trusted_fingerprint = trusted_fingerprint


class OcspError(CertificateVerificationError):
    """Base exception for OCSP errors."""


class OcspRevokedStatusError(OcspError):
    """OCSP response indicated that certificate has been revoked."""

    def __init__(self, certificate: str) -> None:
        super().__init__(f"OCSP status for {certificate} is REVOKED")


class OcspUnknownStatusError(OcspError):
    """OCSP response indicates that certificate validity is unknown."""

    def __init__(self, certificate: str) -> None:
        super().__init__(f"OCSP status for {certificate} is UNKNOWN")


class OcspHttpError(OcspError):
    """OCSP failed due to an HTTP protocol error.

    Args:
        ocsp_url: OCSP endpoint url.
        http_status: HTTP status indicating the type of error.
    """

    def __init__(self, ocsp_url: str, http_status: int) -> None:
        super().__init__(
            f"HTTP failed with status code {http_status} when "
            f" requesting OCSP status to {ocsp_url}",
        )
        self.ocsp_url = ocsp_url
        self.http_status = http_status


class OcspResponseStatusError(OcspError):
    """OCSP response status is not successful.

    Args:
        status: OCSP response status.
    """

    def __init__(self, status: OCSPResponseStatus) -> None:
        super().__init__(f"OCSP response status was {status}")
        self.status = status


class OcspResponseUnsignedError(OcspError):
    """OCSP response is unsigned."""

    def __init__(self) -> None:
        super().__init__("OCSP response is unsigned this should not happen")


class OcspResponseSignatureError(OcspError):
    """OCSP responder certificate did not sign the response."""

    def __init__(self, responder: str) -> None:
        super().__init__(
            f"responder's certificate {responder} did not sign the OCSP response",
        )
        self.responder = responder


class OcspResponderCertificateError(OcspError):
    """OCSP responder certificate not issued or signed by issuer."""

    def __init__(
        self,
        reason: (
            CertificateIssuerNameError
            | CertificateIssuerNotTrustedError
            | CertificateSignatureError
        ),
    ) -> None:
        super().__init__(f"OCSP certificate error: {reason}")
        self.reason = reason


class CertificateChainError(AiaChaserError):
    """Error detected in a certificates chain of trust."""

    def __init__(
        self,
        index: int,
        reason: CertificateVerificationError | OcspError,
    ) -> None:
        super().__init__(f"certificate at index {index}: {reason}")
        self.reason = reason



from __future__ import annotations

import collections
import ssl
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, cast

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import Encoding


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from cryptography.hazmat.primitives import hashes


def certificates_to_der(certificates: list[x509.Certificate]) -> bytes:
    """DER representation of the given certificates.

    Returns:
        A bytes object with the DER content of `certificates`.
    """
    return b"".join(cert.public_bytes(Encoding.DER) for cert in certificates)


def certificates_to_pem(certificates: list[x509.Certificate]) -> str:
    """PEM representation of the given certificates.

    Returns:
        A string with the PEM content of `certificates`.
    """
    return "\n".join(
        cert.public_bytes(Encoding.PEM).decode("ascii") for cert in certificates
    )


def load_ssl_ca_certificates(
    context: ssl.SSLContext | None = None,
    *,
    force_load: bool = True,
) -> list[x509.Certificate]:
    """Load CA certificates available to Python's `ssl`.

    Args:
        context: The SSL context used to get the default certificates.
            If not provided a default context is created with
            `ssl.SSLContext()`.
        force_load: Forcefully load default certificates into the SSL
            context. Certificates in CA path directory are not loaded
            unless they have been used at leas one by the SSL context.

            For more information see
            [`force_load_default_verify_certificates`][aia_chaser.utils.cert_utils.force_load_default_verify_certificates].

    Returns:
        A list with the CA certificates from `context`.
    """
    context = context or ssl.SSLContext()
    context = ssl.SSLContext()
    if force_load:
        force_load_default_verify_certificates(context)

    # Load trusted certificates
    trusted_der = context.get_ca_certs(True)  # noqa: FBT003
    return list(map(x509.load_der_x509_certificate, trusted_der))


def force_load_default_verify_certificates(context: ssl.SSLContext) -> None:
    """Forcefully load default verify certificates into the SSL context.

    Certificates in CA path directory are not loaded unless they
    have been used at leas one by the SSL context.

    This function loads all files located CA path, except those
    that are considered hidden files (start with a '.').

    Args:
        context: The SSL context to load the verify certificates into.
    """
    # Expected way to load default certificates
    context.load_default_certs(purpose=ssl.Purpose.SERVER_AUTH)

    # Forcefully load files in CA path (this solution may not work on Windows
    ssl_defaults = ssl.get_default_verify_paths()

    if ssl_defaults.capath is not None:
        ca_files = filter(
            lambda ca_file: ca_file.is_file() and not ca_file.name.startswith("."),
            Path(ssl_defaults.capath).iterdir(),
        )
        # Note: We found that some ssl installations have files that
        # are not cert, i.e., homebrew's openssl has a .keepme. For now
        # we ignore files that start with '.' but a try except might
        # be necessary ¯\_(ツ)_/¯
        for ca_file in ca_files:
            context.load_verify_locations(ca_file)


def find_leaf_certificates(
    certificates: Sequence[x509.Certificate],
) -> list[x509.Certificate]:
    """Finds leaf certificates.

    A certificate is considered a leaf certificate if its subject
    is not found as the issuer of another certificate from the list.

    Returns:
        List with the certificates that are not issuers of any of the
            other provided certificates.
    """
    graph = collections.defaultdict(list)
    for cert in certificates:
        graph[cert.issuer].append(cert.subject)

    return [
        cert
        for cert in certificates
        if cert.subject not in graph or not graph[cert.subject]
    ]


def build_certificate_chain(
    leaf_cert: x509.Certificate,
    certs_map: Mapping[x509.Name, x509.Certificate],
) -> list[x509.Certificate]:
    """Builds a certificate chain from the `leaf_cert` to the root CA.

    Args:
        leaf_cert: Leaf certificate of the chain.
        certs_map: Mapping from `x509.Certificate.subject` to `x509.Certificate`.

    Returns:
        A certificate chain starting at `leaf_cert` and ending in root CA.

    Raises:
        KeyError: An issuer is not found in `certs_map`.
    """
    chain = [leaf_cert]

    while chain[-1].subject != chain[-1].issuer:
        chain.append(certs_map[chain[-1].issuer])

    return chain


def build_certificate_chains(
    certificates: Sequence[x509.Certificate],
) -> list[list[x509.Certificate]]:
    """Builds all certificate chains found in `certificates`.

    First it looks for all leaf certificates using
    [`find_leaf_certificates`][aia_chaser.utils.cert_utils.find_leaf_certificates]
    and then builds the chains starting at each leaf using
    [`build_certificate_chain`][aia_chaser.utils.cert_utils.build_certificate_chain].

    Returns:
        All certificate chains found in `certificates` each starting at
            its corresponding leaf certificate.
    """
    leaves = find_leaf_certificates(certificates)
    certs_map = {cert.subject: cert for cert in certificates}

    return [
        build_certificate_chain(
            leaf_cert=leaf,
            certs_map=certs_map,
        )
        for leaf in leaves
    ]


class AiaInformation(NamedTuple):
    """Authority Information Access (AIA) values."""

    ca_issuers: list[str]
    ocsp_urls: list[str]


def extract_aia_information(
    certificate: x509.Certificate,
) -> AiaInformation:
    """Extract authority information access (AIA) from a certificate.

    Args:
        certificate: Certificate from which extract AIA information.

    Returns:
        The extracted CA issues and OCSP servers.

    Note:
        If the certificate does not have the AIA extension this function
        does not fail, it fallbacks to returning empty sequences of data.
    """
    try:
        aia_extension = cast(
            x509.Extension[x509.AuthorityInformationAccess],
            certificate.extensions.get_extension_for_oid(
                x509.ExtensionOID.AUTHORITY_INFORMATION_ACCESS,
            ),
        )
    except x509.ExtensionNotFound:
        return AiaInformation([], [])

    ca_issuers = [
        aia_entry.access_location.value
        for aia_entry in aia_extension.value
        if aia_entry.access_method == x509.AuthorityInformationAccessOID.CA_ISSUERS
    ]
    ocsp_urls = [
        aia_entry.access_location.value
        for aia_entry in aia_extension.value
        if aia_entry.access_method == x509.AuthorityInformationAccessOID.OCSP
    ]

    return AiaInformation(ca_issuers=ca_issuers, ocsp_urls=ocsp_urls)


_PADDING_PKCS1V15_OIDS = (
    x509.SignatureAlgorithmOID.RSA_WITH_MD5,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA1,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA224,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA256,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA384,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA3_224,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA3_256,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA3_384,
    x509.SignatureAlgorithmOID.RSA_WITH_SHA3_512,
)


def select_rsa_padding_for_signature_algorithm_oid(
    signature_alg_oid: x509.ObjectIdentifier,
    signature_hash_alg: hashes.HashAlgorithm | None,
) -> padding.AsymmetricPadding:
    """Select padding for a given signature algorithm OID."""
    if signature_alg_oid == x509.SignatureAlgorithmOID.RSASSA_PSS:
        if signature_hash_alg is None:
            msg = "RSASSA-PSS signature requires a hash algorithm"
            raise ValueError(msg)

        return padding.PSS(
            mgf=padding.MGF1(signature_hash_alg),
            salt_length=padding.PSS.MAX_LENGTH,
        )
    if signature_alg_oid in _PADDING_PKCS1V15_OIDS:
        return padding.PKCS1v15()

    msg = f"unknown padding for signature algorithm oid {signature_alg_oid}"
    raise ValueError(msg)



from types import MappingProxyType

from aia_chaser.utils.type_utils import StrEnum


DOWNLOAD_CACHE_SIZE = 1024

DEFAULT_URLOPEN_TIMEOUT = 30

DEFAULT_SCHEME_PORT = MappingProxyType(
    {
        "dns": 53,
        "ftp": 21,
        "http": 80,
        "https": 443,
        "ipp": 631,
        "nfs": 111,
        "ssh": 22,
        "telnet": 23,
    },
)

X509_CERTIFICATE_MIME = (
    "application/x-x509-ca-cert",
    "application/pkix-cert",
    "application/octet-stream",
    "binary/octet-stream",
)


class HttpHeader(StrEnum):
    """HTTP header name."""

    CONTENT_TYPE = "Content-Type"


class SSLReason(StrEnum):
    """SSL error reasons."""

    NO_CERTIFICATE_FOUND = "NO_CERTIFICATE_OR_CRL_FOUND"



import enum
import sys


if sys.version_info >= (3, 11):
    StrEnum = enum.StrEnum
else:

    class StrEnum(str, enum.Enum):
        """Enum subclass to define string constants that behave like str.

        Additionally it also redefines __str__ to return the value instead of
        showing a string that also includes the Enum's part.
        """

        def __str__(self) -> str:
            """String representation of the str-based enum object.

            Returns:
                The value of the enum object, which is already an str.
            """
            return self.value



from __future__ import annotations

import contextlib
import dataclasses
import datetime
import http
import itertools
import secrets
import warnings
from typing import TYPE_CHECKING
from urllib.request import Request, urlopen

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.x509 import ocsp

from aia_chaser.exceptions import (
    CertificateChainError,
    CertificateExpiredError,
    CertificateFingerprintError,
    CertificateIssuerNameError,
    CertificateIssuerNotTrustedError,
    CertificateKeyTypeError,
    CertificateSignatureError,
    CertificateTimeZoneError,
    CertificateVerificationError,
    OcspError,
    OcspHttpError,
    OcspResponderCertificateError,
    OcspResponseSignatureError,
    OcspResponseStatusError,
    OcspResponseUnsignedError,
    OcspRevokedStatusError,
    OcspUnknownStatusError,
    RootCertificateNotFoundError,
)
from aia_chaser.utils.cert_utils import (
    extract_aia_information,
    select_rsa_padding_for_signature_algorithm_oid,
)


if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping


def _get_default_verification_time() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


@dataclasses.dataclass
class VerifyCertificatesConfig:
    """Configuration to verify certificates.

    Attributes:
        fingerprint_hash_alg: Hash algorithm used to verify that the
            root certificate from the chain is the same as the one
            found in the trusted certificates. Defaults to `hashes.SHA256`.

        ocsp_enabled: Whether or not perform OCSP validation.
            Defaults to True.
        ocsp_hash_alg: Hash algorithm used to construct OCSP requests.
            Defaults to `hashes.SHA1`.
        ocsp_ignore_unknown: Whether to ignore OCSP's status UNKNOWN or
            consider it a verification error.
            Defaults to True.
        ocsp_verify_responder: Whether to verify responder certificate. See
            [`VerifyOcspConfig`][aia_chaser.verify.VerifyOcspConfig] for more
            details.


        verification_time: Timestamp used to verify certificate validity
            period. Defaults to `datetime.datetime.now(datetime.timezone.utc)`.
    """

    fingerprint_hash_alg: hashes.HashAlgorithm = dataclasses.field(
        default_factory=hashes.SHA256,
    )

    ocsp_enabled: bool = False
    ocsp_hash_alg: hashes.HashAlgorithm = dataclasses.field(
        default_factory=hashes.SHA1,
    )
    ocsp_ignore_unknown: bool = True
    ocsp_verify_responder: bool = True

    verification_time: datetime.datetime = dataclasses.field(
        default_factory=_get_default_verification_time,
    )


def verify_certificate_chain(
    certificates: Iterable[x509.Certificate],
    trusted: Mapping[x509.Name, x509.Certificate] | None = None,
    config: VerifyCertificatesConfig | None = None,
) -> None:
    """Verifies the integrity of the certificates chain.

    The verification checks that each certificate in the sequence
    is signed by the next one and that all are valid certificates.

    Args:
        certificates: Chain of certificates starting with the leaf and
            ending in the root CA certificate.
        trusted: Trusted certificates mapping from subject, formatted as
            rfc4514, to certificate. If not provided root certificate
            verification will be skipped.
        config: Configuration of the verification process.

    Raises:
        CertificateChainError: If a verification error is detected on
            any of the certificates from the chain. It will also
            be raised if trusted is given `certificate[-1]` is not
            contained in it or the fingerprint does not match.
    """
    config = config or VerifyCertificatesConfig()

    with contextlib.suppress(StopIteration):
        certificates, ca_certificates = itertools.tee(certificates)
        root_cert = next(ca_certificates)

    chain_index = 0
    try:
        for issued, issuer in zip(certificates, ca_certificates):
            root_cert = issuer

            verify_certificate_validity_period(
                issued,
                verification_time=config.verification_time,
            )
            verify_directly_issued_by(certificate=issued, issuer=issuer)
            if config.ocsp_enabled:
                verify_ocsp_status(
                    certificate=issued,
                    issuer=issuer,
                    config=VerifyOcspConfig(
                        hash_alg=config.ocsp_hash_alg,
                        ignore_unknown=config.ocsp_ignore_unknown,
                        verify_responder=config.ocsp_verify_responder,
                        trusted=trusted or {},
                    ),
                )

            chain_index += 1

        if trusted is not None:
            verify_root_certificate(
                root_cert,
                hash_alg=config.fingerprint_hash_alg,
                trusted=trusted,
                verification_time=config.verification_time,
            )
    except (CertificateVerificationError, OcspError) as err:
        raise CertificateChainError(index=chain_index, reason=err) from err


def verify_directly_issued_by(
    certificate: x509.Certificate,
    issuer: x509.Certificate,
) -> None:
    """Verifies that a certificate was issued by the provided issuer.

    This function delegates to `x509.Certificate.verify_directly_issued_by`
    to check if the given certificate's issuer matches the provided issuer
    certificate and to validate the issuer's signature. If the check fails,
    specific exceptions are raised to indicate the type of failure.

    Args:
        certificate: The certificate to validate.
        issuer: The certificate of the issuer expected to have issued the
            given certificate.

    Raises:
        CertificateIssuerNameError: If the issuer's subject name does not match
            the certificate's issuer name.
        CertificateSignatureError: If the issuer's signature on the certificate
            is invalid.
    """
    try:
        certificate.verify_directly_issued_by(issuer)
    except TypeError as err:
        raise CertificateKeyTypeError(str(err)) from err
    except ValueError as err:
        raise CertificateIssuerNameError(
            cert_issuer=certificate.subject.rfc4514_string(),
            issuer_subject=issuer.subject.rfc4514_string(),
        ) from err
    except InvalidSignature as err:
        raise CertificateSignatureError(
            certificate=certificate.subject.rfc4514_string(),
            issuer=issuer.subject.rfc4514_string(),
        ) from err


def verify_root_certificate(
    root_cert: x509.Certificate,
    trusted: Mapping[x509.Name, x509.Certificate],
    verification_time: datetime.datetime | None = None,
    hash_alg: hashes.HashAlgorithm | None = None,
) -> None:
    """Verifies the validity of the provided root certificate.

    Args:
        root_cert: Certificate to verify.
        verification_time: datetime value to validate the certificates
            validity period.
            Defaults to `datetime.datetime.now(datetime.timezone.utc)`.
        trusted: Trusted certificates mapping from subject (name) to
            certificate.
        hash_alg: Hashing algorithm used for operations like fingerprint
            comparison, etc. Defaults to SHA-256.

    Raises:
        RootCertificateNotFoundError: If the root certificate cannot
            be found in `trusted`.
        CertificateFingerprintError: If the root certificate fingerprint
            does not match the trusted certificate fingerprint.
    """
    verification_time = verification_time or _get_default_verification_time()
    hash_alg = hash_alg or hashes.SHA256()

    if root_cert.subject not in trusted:
        raise RootCertificateNotFoundError(root_cert.subject.rfc4514_string())

    # Match fingerprint
    trusted_root_ca = trusted[root_cert.subject]
    trusted_root_cert_fp = trusted_root_ca.fingerprint(hash_alg)
    root_cert_fp = root_cert.fingerprint(hash_alg)

    if trusted_root_cert_fp != root_cert_fp:
        raise CertificateFingerprintError(
            fingerprint=root_cert_fp,
            trusted_fingerprint=trusted_root_cert_fp,
        )

    # Verify validity period
    verify_certificate_validity_period(
        root_cert,
        verification_time=verification_time,
    )


def verify_certificate_validity_period(
    certificate: x509.Certificate,
    verification_time: datetime.datetime | None = None,
) -> None:
    """Verifies certificate validity period (not valid before/after).

    Args:
        certificate: Certificate to verify.
        verification_time: datetime value to use as reference when
            verifying the validity period. If not given uses UTC time.

    Raises:
        CertificateExpiredError: If the certificate is outside its validity
            period.
        CertificateTimeZoneError: If `verification_time` is offset-naive.
    """
    verification_time = verification_time or _get_default_verification_time()

    if verification_time.tzinfo is None:
        raise CertificateTimeZoneError

    not_valid_before = _get_not_valid_before(certificate)
    not_valid_after = _get_not_valid_after(certificate)

    if not not_valid_before <= verification_time <= not_valid_after:
        raise CertificateExpiredError(
            not_valid_before=not_valid_before,
            not_valid_after=not_valid_after,
            verification_time=verification_time,
        )


######################
# Verify OCSP Status #
######################


@dataclasses.dataclass
class VerifyOcspConfig:
    """Configuration to verify certificates.

    Attributes:
        hash_alg: Hash algorithm used to construct OCSP requests.
            Defaults to `hashes.SHA1`.
        nonce_size: Size in bytes of the OCSP nonce.
        ignore_unknown: Whether to ignore OCSP's status UNKNOWN or
            consider it a verification error.
            Defaults to True.

        verify_responder: Whether to verify responder certificate. If enabled
            there are 3 possibilities:

                1 - Responder and Issuer are the same: Issuer is *trusted*
                    (verifying it should be done by the user using
                    `verify_certificate_chain`) and the response signature
                    is verified with the Issuer certificate.
                2 - Responder certificate is different but is signed by
                    Issuer: The same criteria applies, Issuer is *trusted*
                    and used to verify the responder certificate, which
                    in turn is used to verify the response.
                3 - Responder certificate is different and signed by another
                    certificate: Then a trusted set of certificates is
                    required to verify the certificate.

        trusted: Trusted certificates mapping from subject (name) to
            certificate. If not provided a responder certificate
            verification may fail.
    """

    hash_alg: hashes.HashAlgorithm = dataclasses.field(
        default_factory=hashes.SHA1,
    )
    nonce_size: int = 20
    ignore_unknown: bool = True
    verify_responder: bool = False

    trusted: Mapping[x509.Name, x509.Certificate] = dataclasses.field(
        default_factory=dict,
    )

    def __post_init__(self) -> None:  # noqa: D105
        if self.nonce_size < 1:
            raise ValueError("`nonce_size` must be at least 1")  # noqa: EM101, TRY003


def verify_ocsp_status(
    certificate: x509.Certificate,
    issuer: x509.Certificate,
    config: VerifyOcspConfig | None = None,
) -> None:
    """Verifies the status of a certificate using Online Certificate Status Protocol.

    Args:
        certificate: The certificate whose revocation status needs to be
            verified.
        issuer: The issuer certificate that signed the target certificate.
        config: Configuration of the OCSP verification process.

    Raises:
        OcspRevokedStatusError:
            The certificate status is `REVOKED` in the OCSP response.
        OcspUnknownStatusError:
            The certificate status is `UNKNOWN` in the OCSP response
            and `ignore_unknown` is False.
        OcspHttpError:
            An HTTP error happened while requesting the OCSP status.
        OcspResponseStatusError:
            The OCSP response status is not `SUCCESSFUL`.
        OcspResponseUnsignedError:
            The OCSP response is not signed.
        OcspResponseSignatureError:
            The OCSP response is not signed by the responder certificate.
        CertificateKeyTypeError:
            The responder certificate key is not supported.

    Warnings:
        UserWarning:
            A warning is issued if a hash algorithm other than SHA-1 is
            used, as some OCSP servers may not support other hash algorithms
            and may fail with `UNAUTHORIZED` or `MALFORMED` responses.
    """
    config = config or VerifyOcspConfig()

    aia_info = extract_aia_information(certificate)
    nonce = secrets.token_bytes(config.nonce_size)

    if config.hash_alg.name.lower() != "sha1":
        warnings.warn(
            "Some OCSP servers may fail with UNAUTHORIZED or MALFORMED "
            " responses if request hash is different than SHA1",
            category=UserWarning,
            stacklevel=2,
        )

    if aia_info.ocsp_urls:
        builder = ocsp.OCSPRequestBuilder()
        builder = builder.add_certificate(certificate, issuer, config.hash_alg)
        builder = builder.add_extension(x509.OCSPNonce(nonce), critical=False)
        ocsp_request = builder.build()

        for ocsp_url in aia_info.ocsp_urls:
            last_status = _run_ocsp_request(
                ocsp_request=ocsp_request,
                ocsp_url=ocsp_url,
                issuer=issuer,
                config=config,
            )
            if last_status == ocsp.OCSPCertStatus.REVOKED:
                raise OcspRevokedStatusError(certificate.subject.rfc4514_string())
            if last_status == ocsp.OCSPCertStatus.UNKNOWN and not config.ignore_unknown:
                raise OcspUnknownStatusError(certificate.subject.rfc4514_string())


def _run_ocsp_request(
    ocsp_request: ocsp.OCSPRequest,
    ocsp_url: str,
    issuer: x509.Certificate,
    config: VerifyOcspConfig,
) -> ocsp.OCSPCertStatus:
    http_request = Request(  # noqa: S310
        url=ocsp_url,
        data=ocsp_request.public_bytes(serialization.Encoding.DER),
        headers={"Content-Type": "application/ocsp-request"},
        method="POST",
    )

    with urlopen(http_request) as response:  # noqa: S310
        if response.status != http.HTTPStatus.OK:
            raise OcspHttpError(ocsp_url=ocsp_url, http_status=response.status)
        ocsp_response_data = response.read()

    ocsp_resp = ocsp.load_der_ocsp_response(ocsp_response_data)
    if ocsp_resp.response_status != ocsp.OCSPResponseStatus.SUCCESSFUL:
        raise OcspResponseStatusError(ocsp_resp.response_status)

    all_responses = list(ocsp_resp.responses)
    if len(all_responses) != 1:
        raise OcspError("exactly one response is expected")  # noqa: EM101, TRY003
    single_response = all_responses[0]

    # Verify response signature
    if ocsp_resp.signature_hash_algorithm is None:
        raise OcspResponseUnsignedError

    responder_cert = _get_ocsp_responder_certificate(
        ocsp_response=ocsp_resp,
        issuer=issuer,
    )
    try:
        _verify_ocsp_responder_certificate(
            responder_cert=responder_cert,
            issuer=issuer,
            config=config,
        )
    except (
        CertificateIssuerNameError,
        CertificateIssuerNotTrustedError,
        CertificateSignatureError,
    ) as err:
        raise OcspResponderCertificateError(err) from err

    responder_key = responder_cert.public_key()

    try:
        if isinstance(responder_key, rsa.RSAPublicKey):
            responder_key.verify(
                signature=ocsp_resp.signature,
                data=ocsp_resp.tbs_response_bytes,
                padding=select_rsa_padding_for_signature_algorithm_oid(
                    signature_alg_oid=ocsp_resp.signature_algorithm_oid,
                    signature_hash_alg=ocsp_resp.signature_hash_algorithm,
                ),
                algorithm=ocsp_resp.signature_hash_algorithm,
            )
        elif isinstance(responder_key, ec.EllipticCurvePublicKey):
            responder_key.verify(
                signature=ocsp_resp.signature,
                data=ocsp_resp.tbs_response_bytes,
                signature_algorithm=ec.ECDSA(
                    ocsp_resp.signature_hash_algorithm,
                ),
            )
        else:
            raise CertificateKeyTypeError(
                reason=f"unsupported key type {type(responder_key)}",
            )
    except InvalidSignature:
        raise OcspResponseSignatureError(
            responder_cert.subject.rfc4514_string(),
        ) from None

    # Check certificate status
    return single_response.certificate_status


def _get_ocsp_responder_certificate(
    ocsp_response: ocsp.OCSPResponse,
    issuer: x509.Certificate,
) -> x509.Certificate:
    # We only request OCSP of a single certificate so the response
    # should only contain one responder certificate, if any.
    if ocsp_response.certificates:
        return ocsp_response.certificates[0]

    # Fallback to issuer as responder
    return issuer


def _verify_ocsp_responder_certificate(
    responder_cert: x509.Certificate,
    issuer: x509.Certificate,
    config: VerifyOcspConfig,
) -> None:
    if config.verify_responder and responder_cert != issuer:
        responder_issuer = (
            issuer
            if responder_cert.issuer == issuer.subject
            else config.trusted.get(responder_cert.issuer)
        )
        if responder_issuer is not None:
            verify_directly_issued_by(
                certificate=responder_cert,
                issuer=responder_issuer,
            )
        else:
            raise CertificateIssuerNotTrustedError(
                issuer=responder_cert.issuer.rfc4514_string(),
            )


# Since version 42 not_valid_before and not_valid_after are deprecated
# in favor of the offset-aware alternatives not_valid_before_utc and
# not_valid_after_utc.
if hasattr(x509.Certificate, "not_valid_before_utc"):

    def _get_not_valid_before(cert: x509.Certificate) -> datetime.datetime:
        return cert.not_valid_before_utc

    def _get_not_valid_after(cert: x509.Certificate) -> datetime.datetime:
        return cert.not_valid_after_utc

else:

    def _get_not_valid_before(cert: x509.Certificate) -> datetime.datetime:
        return cert.not_valid_before.replace(tzinfo=datetime.timezone.utc)

    def _get_not_valid_after(cert: x509.Certificate) -> datetime.datetime:
        return cert.not_valid_after.replace(tzinfo=datetime.timezone.utc)
